// sndstretch.h
//
//    sndstretch - algorithm for adjusting pitch and speed of s16le data
//    Copyright (C) 2000  Florian Berger
//    Email: florian.berger@jk.uni-linz.ac.at
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License Version 2 as
//    published by the Free Software Foundation;
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//

#ifndef SNDSTRETCH_H
#define SNDSTRETCH_H


#ifdef __cplusplus
extern "C" {
#endif


typedef short int s16;


int sndstretch(
// stretches the sound (not changing pitch !!!)
// returns number of output samples produced
           s16 * buffer, int buff_size,   // ring buffer
           int pos_init,                  // only initial pos in ringbuffer
           int snr_i, int snr_o,          // enlarge - spec
           int chnr,                      /* # of channels                  */
           s16 * outbuff,                 // output
           int * out_prod,                // # of output-samples produced
           int snr_proc,                  // # of in-samples to process
           int initialize                 // bool
              );

    
int sndscale(
// rescales the sound (including pitch)
// returns number of output samples produced
           s16 * buffer,                  // ring buffer
           int snr_i, int snr_o,          // enlarge - spec
           int chnr,                      // # of channels
           s16 * outbuff,                 // output
           int * out_prod,                // # of output-samples produced
           int snr_proc,                  // # of in-samples to process
           int initialize                 // bool
          );


int snd_stretch_scale(s16 *buff_i, s16 * buff_o,
                      double pitch, double speed, int channels,
                      int snr_proc, int * snr_produced, int initialize );

#ifdef __cplusplus
}
#endif

#endif // SNDSTRETCH_H
