// sndstretch_xmms.c
//
//    sndstretch_xmms - xmms-output plugin for adjusting
//                      pitch and speed of s16le data
//    Copyright (C) 2001  Florian Berger
//    Email: florian.berger@jk.uni-linz.ac.at
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License Version 2 as
//    published by the Free Software Foundation;
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//

#include <gtk/gtk.h>
#include "sndstretch_xmms.h"
//#include "sndstretch/sndstretch.h"
#include "sndstretch.h"
#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/soundcard.h>
#include <unistd.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>

#define SS sndstretch_var

OutputPlugin sndstretch_op = {
	NULL,		/* Filled in by xmms */
	NULL,		/* Filled in by xmms */
	"SndStretch - sound stretcher ",	/* The description that is shown in the preferences box */
        sndstretch_init,     //	   NULL,           /* void *init(void) */
	sndstretch_about,       /* Show the about box */
	sndstretch_config,      /* Show the configuration dialog */
	sndstretch_get_volume,
	sndstretch_set_volume,
	sndstretch_open_audio,	/* Open the device, if the device can't handle the given
								   parameters the plugin is responsible for downmixing
								   the data to the right format before outputting it */
	sndstretch_write_audio,	/* The input plugin calls this to write data to the output
							   buffer */
	sndstretch_close_audio,	/* No comment... */
        sndstretch_flush,       /* Flush the buffer and set the plugins internal timers to time */
	sndstretch_pause,	/* Pause or unpause the output */
	sndstretch_buffer_free,	/* Return the amount of data that can be written to the buffer,
					   two calls to this without a call to write_audio should make
					   the plugin output audio directly */
	sndstretch_buffer_playing,	/* Returns TRUE if the plugin currently is playing some audio,
					   otherwise return FALSE */
	sndstretch_output_time, 	/* Return the current playing time */
	sndstretch_written_time 	/* Return the length of all the data that has been written to buffer */
};


OutputPlugin *get_oplugin_info(void)
{
	return &sndstretch_op;
}



static struct {
    int handle;    // file handle
    int fragsize;
    int chnr;
    int paused;
    int time_offs;
    int fmtsize;
    int fmt;
    int sampfreq;
    int written;
    int bpsec;
    int vol_l,vol_r;
    int going;
    double pitch;
    double speed;
    GtkObject * pitch_adj;
    GtkObject * speed_adj;
} sndstretch_var;


static const char sndstretch_about_text[] = "Sound Stretcher Plugin\n\n \
By Florian Berger\n<florian.berger@jk.uni-linz.ac.at> \n\n (c) 2001";

GtkWidget * sndstretch_config_dialog = NULL;
GtkWidget * sndstretch_about_dialog  = NULL;


static gint sndstretch_about_destroy_cb(GtkWidget * w, GdkEventAny * e, gpointer data)
{
	gtk_widget_destroy(sndstretch_about_dialog);
	sndstretch_about_dialog = NULL;
	return TRUE;
}

static void sndstretch_about_ok_cb(GtkButton * button, gpointer data)
{
	gtk_widget_destroy(GTK_WIDGET(sndstretch_about_dialog));
	sndstretch_about_dialog = NULL;
}

void sndstretch_about(void)
{
	GtkWidget *hbox, *text, *button;

	if (sndstretch_about_dialog != NULL)
		return;

	sndstretch_about_dialog = gtk_dialog_new();
	gtk_signal_connect(GTK_OBJECT(sndstretch_about_dialog), "destroy",
			   GTK_SIGNAL_FUNC(sndstretch_about_destroy_cb), NULL);
	gtk_window_set_title(GTK_WINDOW(sndstretch_about_dialog), "About SoundStretcher");

	hbox = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(sndstretch_about_dialog)->vbox), hbox,
			   TRUE, TRUE, 5);
	gtk_widget_show(hbox);

	text = gtk_text_new(NULL, NULL);
	gtk_text_set_editable(GTK_TEXT(text), FALSE);
	gtk_text_insert(GTK_TEXT(text), NULL, NULL, NULL,
			sndstretch_about_text, strlen(sndstretch_about_text));
	gtk_box_pack_start(GTK_BOX(hbox), text, TRUE, TRUE, 5);
	gtk_widget_show(text);

	button = gtk_button_new_with_label("Ok");
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(sndstretch_about_dialog)->action_area),
			   button, TRUE, TRUE, 0);
	gtk_signal_connect(GTK_OBJECT(button), "clicked",
			   GTK_SIGNAL_FUNC(sndstretch_about_ok_cb), NULL);
	GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
	gtk_widget_grab_default(button);
	gtk_widget_show(button);
        gtk_widget_set_usize (sndstretch_about_dialog,250,160);
	gtk_widget_show(sndstretch_about_dialog);
}



static gint sndstretch_config_destroy_cb(GtkWidget * w, GdkEventAny * e, gpointer data)
{
	gtk_widget_destroy(sndstretch_config_dialog);
	sndstretch_config_dialog = NULL;
	return TRUE;
}

static void speed_change_cb(GtkAdjustment * adj, gpointer data)
{
    SS.speed = pow( 2.0, GTK_ADJUSTMENT(adj)->value / (GTK_ADJUSTMENT(adj)->upper-10) );
}

static void pitch_change_cb(GtkAdjustment * adj, gpointer data)
{
    SS.pitch = pow( 2.0, GTK_ADJUSTMENT(adj)->value / (GTK_ADJUSTMENT(adj)->upper-10) );
}

void sndstretch_config(void)
{
    GtkWidget *vbox;
    GtkWidget *speed_scale, *pitch_scale;
    GtkWidget *speed_spin,  *pitch_spin;
    GtkWidget *speed_hbox,  *pitch_hbox;
    GtkWidget *speed_frame, *pitch_frame;

	if (sndstretch_config_dialog != NULL)
		return;

        SS.speed_adj = gtk_adjustment_new(0.0, -100, 100+10, 2, 10, 10);
        SS.pitch_adj = gtk_adjustment_new(0.0, -120, 120+10, 2, 10, 10);

	gtk_signal_connect(GTK_OBJECT(SS.speed_adj), "value-changed",
			   GTK_SIGNAL_FUNC(speed_change_cb), NULL);

	gtk_signal_connect(GTK_OBJECT(SS.pitch_adj), "value-changed",
                           GTK_SIGNAL_FUNC(pitch_change_cb), NULL);

        speed_scale = gtk_hscale_new(GTK_ADJUSTMENT(SS.speed_adj));
        pitch_scale = gtk_hscale_new(GTK_ADJUSTMENT(SS.pitch_adj));
        gtk_scale_set_draw_value (GTK_SCALE(speed_scale),FALSE);
        gtk_scale_set_draw_value (GTK_SCALE(pitch_scale),FALSE);

        speed_spin = gtk_spin_button_new(GTK_ADJUSTMENT(SS.speed_adj),1.0,2);
        pitch_spin = gtk_spin_button_new(GTK_ADJUSTMENT(SS.pitch_adj),1.0,2);
        gtk_widget_set_usize (speed_spin,70,20);
        gtk_widget_set_usize (pitch_spin,70,20);
        gtk_entry_set_max_length (GTK_ENTRY(pitch_spin),7);
        gtk_entry_set_max_length (GTK_ENTRY(speed_spin),7);

        speed_hbox = gtk_hbox_new(FALSE,5);
        pitch_hbox = gtk_hbox_new(FALSE,5);
        gtk_container_set_border_width(GTK_CONTAINER(speed_hbox), 3);
        gtk_container_set_border_width(GTK_CONTAINER(pitch_hbox), 3);
        gtk_box_pack_start(GTK_BOX(speed_hbox), speed_spin,  FALSE, TRUE, 5);
        gtk_box_pack_start(GTK_BOX(speed_hbox), speed_scale, TRUE, TRUE, 5);
        gtk_box_pack_start(GTK_BOX(pitch_hbox), pitch_spin,  FALSE, TRUE, 5);
        gtk_box_pack_start(GTK_BOX(pitch_hbox), pitch_scale, TRUE, TRUE, 5);

        speed_frame = gtk_frame_new("Speed");
        pitch_frame = gtk_frame_new("Pitch");
        gtk_container_add(GTK_CONTAINER(speed_frame), speed_hbox);
        gtk_container_add(GTK_CONTAINER(pitch_frame), pitch_hbox);
        gtk_container_set_border_width(GTK_CONTAINER(speed_frame), 5);
        gtk_container_set_border_width(GTK_CONTAINER(pitch_frame), 5);

        vbox=gtk_vbox_new(FALSE,0);
        gtk_box_pack_start(GTK_BOX(vbox), pitch_frame, FALSE, TRUE, 0);
        gtk_box_pack_start(GTK_BOX(vbox), speed_frame, FALSE, TRUE, 0);

	sndstretch_config_dialog = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_signal_connect(GTK_OBJECT(sndstretch_config_dialog), "destroy",
			   GTK_SIGNAL_FUNC(sndstretch_config_destroy_cb), NULL);
	gtk_window_set_title(GTK_WINDOW(sndstretch_config_dialog), "SoundStretcher - configuartion");
        gtk_container_add(GTK_CONTAINER(sndstretch_config_dialog), vbox);

        gtk_widget_set_usize (sndstretch_config_dialog,275,116);

//	gtk_widget_realize(sndstretch_config_dialog);
//        gdk_window_set_decorations(sndstretch_config_dialog->window, 0);

        gtk_widget_show_all(sndstretch_config_dialog);
}



void init_audio(
                int *aud_handle,      // io handle
                int *fragsize,        // fragment size
                int sampfreq,         // sample frequency
                int stereo,           // 0=mono, 1=stereo
                int format,           // sample format
                const char *device,   // read write device
                int read_access,      // read or write
                int fraglog,          // log2(fragsize)
                int fragnr            // # of frags in buffer
               )
    //  initialize the soundcard
{

   int audio_io;
   int blksize;
   int a_sampsize;
   int a_stereo;
   int a_speed;
//   int mixer_i,mixer_o;
//   int vol;
   audio_buf_info info;
   int fmts;


   audio_io = open(device, read_access?O_RDONLY:O_WRONLY, 0);


#ifdef DEBUG
   fprintf(stderr,"********************* BEGIN *********************\n");

   fprintf(stderr,"Opened device:%s\n", device);
#endif

#ifdef DEBUG
   {int caps;
       ioctl( audio_io, SNDCTL_DSP_GETCAPS, &caps );
       fprintf(stderr,"OSS-Version %d\n",caps & DSP_CAP_REVISION);
       fprintf(stderr,"  DUPLEX   = %X\n",caps & DSP_CAP_DUPLEX   );
       fprintf(stderr,"  REALTIME = %X\n",caps & DSP_CAP_REALTIME );
       fprintf(stderr,"  BATCH    = %X\n",caps & DSP_CAP_BATCH    );
       fprintf(stderr,"  COPROC   = %X\n",caps & DSP_CAP_COPROC   );
       fprintf(stderr,"  TRIGGER  = %X\n",caps & DSP_CAP_TRIGGER  );
       fprintf(stderr,"  MMAP     = %X\n",caps & DSP_CAP_MMAP     );
   }
#endif
   
//   blksize=0x00030006;  // ( 2^4 = 16 bytes ) * ( 3 fragments )
//   ioctl(audio_i, SNDCTL_DSP_SETFRAGMENT, &blksize);
   
//   blksize=0x0008000A;  // 2^4 = 16
//   blksize=0x00080004;  // 2^4 = 16
   blksize = (fragnr<<16) + fraglog;

//   fprintf(stderr,"   blksize  = %0X\n",blksize );
   
//   blksize=0x00000004;  // 4 * 2^4 = 64 byte
   ioctl(audio_io, SNDCTL_DSP_SETFRAGMENT, &blksize);
   
//ioctl(audio, SNDCTL_DSP_SETBLKSIZE, &blksize);
   
   ioctl(audio_io, SNDCTL_DSP_GETOSPACE, &info);
#ifdef DEBUG
   fprintf(stderr,"Info for output\n");
   fprintf(stderr,"   fragments  = %d\n",info.fragments );
   fprintf(stderr,"   fragstotal = %d\n",info.fragstotal);
   fprintf(stderr,"   fragsize   = %d\n",info.fragsize  );
   fprintf(stderr,"   bytes      = %d\n",info.bytes     );
#endif

   *fragsize=info.fragsize;
   
   ioctl(audio_io, SNDCTL_DSP_GETFMTS, &fmts);
#ifdef DEBUG
   fprintf(stderr,"Formats supported (output):\n");
   fprintf(stderr,"   AFMT_QUERY      = %d\n", (AFMT_QUERY      & fmts)!=0);
   fprintf(stderr,"   AFMT_MU_LAW     = %d\n", (AFMT_MU_LAW     & fmts)!=0);
   fprintf(stderr,"   AFMT_A_LAW      = %d\n", (AFMT_A_LAW      & fmts)!=0);
   fprintf(stderr,"   AFMT_IMA_ADPCM  = %d\n", (AFMT_IMA_ADPCM  & fmts)!=0);
   fprintf(stderr,"   AFMT_U8         = %d\n", (AFMT_U8         & fmts)!=0);
   fprintf(stderr,"   AFMT_S16_LE     = %d\n", (AFMT_S16_LE     & fmts)!=0);
   fprintf(stderr,"   AFMT_S8         = %d\n", (AFMT_S8         & fmts)!=0);
   fprintf(stderr,"   AFMT_U16_LE     = %d\n", (AFMT_U16_LE     & fmts)!=0);
   fprintf(stderr,"   AFMT_U16_BE     = %d\n", (AFMT_U16_BE     & fmts)!=0);
   fprintf(stderr,"   AFMT_MPEG       = %d\n", (AFMT_MPEG       & fmts)!=0);
#endif

   if( (fmts & format) == 0 ){
       fprintf(stderr,"!!!! Audioformat not supported by driver/hardware !!!!\n");
   }

   ioctl(audio_io, SOUND_PCM_READ_CHANNELS, &blksize);
#ifdef DEBUG
   fprintf(stderr,"channels=%d\n",blksize);
#endif

//   ioctl(audio_i, SNDCTL_DSP_SYNC, NULL);
//   ioctl(audio_o, SNDCTL_DSP_SYNC, NULL);

//   a_sampsize = AFMT_S16_LE;
   a_sampsize = format;
   ioctl(audio_io, SNDCTL_DSP_SAMPLESIZE, &a_sampsize);
   
   a_stereo = stereo;
   ioctl(audio_io, SNDCTL_DSP_STEREO, &a_stereo);
   
   a_speed = sampfreq;
#ifdef DEBUG
   fprintf(stderr,"setting to:                          %d\n",a_speed);
#endif
   ioctl(audio_io, SNDCTL_DSP_SPEED, &a_speed);
#ifdef DEBUG
   fprintf(stderr,"sampfreq after SNDCTL_DSP_SPEED    = %d\n",a_speed);
#endif
   ioctl(audio_io, SOUND_PCM_READ_RATE, &a_speed);
#ifdef DEBUG
   fprintf(stderr,"SOUND_PCM_READ_RATE (sampfreq got) = %d\n",a_speed);
#endif

#ifdef DEBUG
   fprintf(stderr,"********************** END **********************\n");
#endif
   //sampfreq_exact=sampfreq;

//   mixer_o = open("/dev/mixer", O_RDONLY, 0);
//   vol=0xFFFF;
//   ioctl(mixer_o, SOUND_MIXER_WRITE_VOLUME, &vol);
//   mixer_i = open("/dev/mixer1", O_RDONLY, 0);
//   vol=0xFFFF;
//   ioctl(mixer_i, SOUND_MIXER_WRITE_PCM, &vol);
   
   //printf("MIC-Volume=%d\n",vol);
   //close(mixer);

//   printf("Mixer set\n");

   *aud_handle=audio_io;
}


void sndstretch_init(void)
{
    s16 x[1024];
    s16 y[1024];
    int prod_size;

    SS.fragsize=0;
    SS.chnr=2;
    SS.paused=0;
    SS.time_offs=0;
    SS.fmtsize=2;
    SS.fmt=FMT_S16_NE;
    SS.sampfreq=44100;
    SS.written=0;
    SS.bpsec=176400;
    SS.vol_r=50;
    SS.vol_l=50;
    SS.pitch=1.00;
    SS.speed=1.00;
//    fprintf(stderr,"sndstretch_init() called\n");
    snd_stretch_scale( x, y, SS.pitch, SS.speed, SS.chnr,
                       1024, &prod_size, 0 );
    sndstretch_config();
}

void sndstretch_get_volume     (int *l, int *r)
{
    int mixer_o,vol;
    mixer_o = open("/dev/mixer", O_RDONLY, 0);
    ioctl(mixer_o, SOUND_MIXER_READ_PCM, &vol);
    *l = (vol & 0xFF);
    *r = ((vol>>8) & 0xFF);
//    *l = (vol & 0xFF)      *100/255;
//    *r = ((vol>>8) & 0xFF) *100/255;
    close(mixer_o);
//    *l=SS.vol_l;
//    *r=SS.vol_r;
}


void sndstretch_set_volume     (int l, int r)
/* Set the volume */
{
    int mixer_o,vol;
    int ll,rr;
//    ll=l*0xFF/100;
//    rr=r*0xFF/100;
    ll=l;
    rr=r;
    mixer_o = open("/dev/mixer", O_WRONLY, 0);
    vol=(ll & 0xFF)+((rr & 0xFF)<<8);
    ioctl(mixer_o, SOUND_MIXER_WRITE_PCM, &vol);
    close(mixer_o);
    SS.vol_l=l;
    SS.vol_r=r;
}


int  sndstretch_open_audio     (AFormat fmt, int rate, int nch)
/* Open the device, if the device can't handle the given
   parameters the plugin is responsible for downmixing
   the data to the right format before outputting it */
{
    int handle,fragsize,format;

//    fprintf(stderr,"sndstretch_open_audio() called\n");
    format=AFMT_S16_BE;

    switch (fmt) {
    case   FMT_U8     :   format = AFMT_U8;      SS.fmtsize=1;  break;
    case   FMT_S8     :   format = AFMT_S8;      SS.fmtsize=1;  break;
    case   FMT_U16_LE :   format = AFMT_U16_LE;  SS.fmtsize=2;  break;
    case   FMT_U16_BE :   format = AFMT_U16_BE;  SS.fmtsize=2;  break;
    case   FMT_U16_NE :   format = AFMT_U16_LE;  SS.fmtsize=2;  break;
    case   FMT_S16_LE :   format = AFMT_S16_LE;  SS.fmtsize=2;  break;
    case   FMT_S16_BE :   format = AFMT_S16_BE;  SS.fmtsize=2;  break;
    case   FMT_S16_NE :   format = AFMT_S16_LE;  SS.fmtsize=2;  break;
    }

    init_audio(
                &handle,      // io handle
                &fragsize,    // fragment size
                rate,         // sample frequency
                nch-1,        // 0=mono, 1=stereo
                format,       // sample format
                "/dev/dsp",   // read write device
                0,            // 0=write
                11,           // log2(fragsize)
                10            // # of frags in buffer
              );

    SS.handle=handle;
    SS.fmt=fmt;
    SS.sampfreq=rate;
    SS.chnr=nch;
    SS.bpsec=SS.fmtsize*SS.chnr*rate;

    SS.written=0;
    SS.time_offs=0;
    SS.going=TRUE;

    return handle;
}


void sndstretch_write_audio    (void *ptr, int length)
/* The input plugin calls this to write data to the output buffer */
{
    static short int buff_o[65536];
    int prod_size;
    EffectPlugin *ep;

//    fprintf(stderr,"sndstretch_write_audio(ptr,%d) called\n",length);
    if(SS.going){
/*	ep = get_current_effect_plugin();
        if(effects_enabled() && ep && ep->mod_samples){
            length = ep->mod_samples( ptr, length, SS.fmt, SS.sampfreq, SS.chnr);
        }*/
        snd_stretch_scale( ptr, (s16 *)buff_o, SS.pitch, SS.speed, SS.chnr,
                           length/SS.fmtsize, &prod_size, 0 );
        write(SS.handle,buff_o,prod_size*SS.fmtsize);
        SS.written+=length;
/*    SS.written+=write(SS.handle,ptr,length);*/
    //    SS.time_offs+=(length  * 1000) / SS.bpsec;
        if( SS.written > SS.bpsec*5 ){
            SS.written-=SS.bpsec*5;
            SS.time_offs+=5000;
        }
    }
}


void sndstretch_close_audio    (void)
{
    SS.going=FALSE;
    SS.written=0;
    SS.time_offs=0;
//    fprintf(stderr,"sndstretch_close_audio() called\n");
    close(SS.handle);
}


void sndstretch_flush          (int time)
/* Flush the buffer and set the plugins internal timers to time */
{
    //    flush( SS.handle );
//    fprintf(stderr,"sndstretch_flush(%d) called\n",time);
    SS.written=0;
    SS.time_offs=time;
}


void sndstretch_pause          (short paused)
/* Pause or unpause the output */
{
    SS.paused=paused;
}


int  sndstretch_buffer_free    (void)
/* Return the amount of data that can be written to the buffer,
   two calls to this without a call to write_audio should make
   the plugin output audio directly */
{
/*    audio_buf_info info;
    ioctl(SS.handle, SNDCTL_DSP_GETOSPACE, &info);
    return(info.fragments*info.fragsize*SS.speed-1);*/
    if(SS.paused){
        return(0);
    }else{
        return(8192);
    }
}


int  sndstretch_buffer_playing (void)
/* Returns TRUE if the plugin currently is playing some audio,
   otherwise return FALSE */
{
//    fprintf(stderr,"sndstretch_buffer_playing() called\n");
//    return(!(SS.paused));
    return(TRUE);
}


int  sndstretch_output_time    (void)
/* Return the current playing time */
{
//    fprintf(stderr,"sndstretch_output_time() called - ");
//    fprintf(stderr,"%d returned\n",SS.time_offs + (gint) ((SS.written  * 1000) / SS.bpsec ));
//    fprintf(stderr,"written = %d\n",SS.written);
    return(
           SS.time_offs + (gint) ((SS.written  * 1000) / SS.bpsec )
          );
}


int  sndstretch_written_time   (void)
/* Return the length of all the data that has been written to buffer */
{
//    fprintf(stderr,"sndstretch_output_time() called - ");
//    fprintf(stderr,"%d returned\n",SS.time_offs + (gint) ((SS.written  * 1000) / SS.bpsec ));
    return(
           SS.time_offs + (gint) ((SS.written  * 1000) / SS.bpsec )
          );
}
